import os
import sys
import autopep8
from collections import defaultdict

from google.protobuf.compiler.plugin_pb2 import CodeGeneratorRequest
from google.protobuf.compiler.plugin_pb2 import CodeGeneratorResponse
from google.protobuf import descriptor_pb2
from purerpc.rpc import Cardinality
import itertools

IMPORT_STRINGS = """import purerpc.server
import purerpc.client
from purerpc.rpc import Cardinality, RPCSignature
"""


def get_python_package(proto_name):
    package_name = proto_name[:-len(".proto")]
    return package_name.replace("/", ".") + "_pb2"


def simple_type(type_):
    simple_type = type_.split(".")[-1]
    return simple_type


def get_python_type(proto_name, proto_type):
    if proto_type.startswith("."):
        return get_python_package(proto_name) + "." + simple_type(proto_type)
    else:
        return proto_type


def generate_single_proto(proto_file: descriptor_pb2.FileDescriptorProto,
                          proto_for_entity):
    contents = IMPORT_STRINGS
    contents += "import {}\n".format(get_python_package(proto_file.name))
    for dep_module in proto_file.dependency:
        contents += "import {}\n".format(get_python_package(dep_module))
    for service in proto_file.service:
        if proto_file.package:
            fully_qualified_service_name = proto_file.package + "." + service.name
        else:
            fully_qualified_service_name = service.name
        contents += "\n\nclass {}Servicer(purerpc.server.Servicer):\n".format(service.name)
        for method in service.method:
            contents += "    async def {}(self, input_message{}):\n".format(
                method.name, "s" if method.client_streaming else "")
            contents += "        raise NotImplementedError()\n\n"
        contents += "    @property\n"
        contents += "    def service(self) -> purerpc.server.Service:\n"
        contents += "        service_obj = purerpc.server.Service(\"{}\")\n".format(
            fully_qualified_service_name)
        for method in service.method:
            cardinality = Cardinality.get_cardinality_for(request_stream=method.client_streaming,
                                                          response_stream=method.server_streaming)
            contents += ("        service_obj.add_method(\"{}\", self.{}, "
                         "RPCSignature({}, {}, {}))\n".format(
                             method.name, method.name, cardinality,
                             get_python_type(proto_for_entity[method.input_type],
                                             method.input_type),
                             get_python_type(proto_for_entity[method.output_type],
                                             method.output_type)))
        contents += "        return service_obj\n\n\n"

        contents += "class {}Stub:\n".format(service.name)
        contents += "    def __init__(self, channel):\n"
        contents += "        self._client = purerpc.client.Client(\"{}\", channel)\n".format(
            fully_qualified_service_name)
        for method in service.method:
            cardinality = Cardinality.get_cardinality_for(request_stream=method.client_streaming,
                                                          response_stream=method.server_streaming)
            contents += ("        self.{} = self._client.get_method_stub("
                         "\"{}\", RPCSignature({}, {}, {}))\n".format(
                             method.name, method.name, cardinality,
                             get_python_type(proto_for_entity[method.input_type],
                                             method.input_type),
                             get_python_type(proto_for_entity[method.output_type],
                                             method.output_type)))
        contents += "\n\n"

    return autopep8.fix_code(contents, options={"experimental": True, "max_line_length": 100})


def main():
    request = CodeGeneratorRequest.FromString(sys.stdin.buffer.read())

    files_to_generate = set(request.file_to_generate)

    response = CodeGeneratorResponse()
    proto_for_entity = dict()
    for proto_file in request.proto_file:
        package_name = proto_file.package
        for named_entity in itertools.chain(proto_file.message_type, proto_file.enum_type,
                                            proto_file.service, proto_file.extension):
            if package_name:
                fully_qualified_name = ".".join(["", package_name, named_entity.name])
            else:
                fully_qualified_name = "." + named_entity.name
            proto_for_entity[fully_qualified_name] = proto_file.name
    for proto_file in request.proto_file:
        if proto_file.name in files_to_generate:
            out = response.file.add()
            out.name = proto_file.name.replace('.proto', "_grpc.py")
            out.content = generate_single_proto(proto_file, proto_for_entity)
    sys.stdout.buffer.write(response.SerializeToString())



import curio.meta
from .grpclib.exceptions import ProtocolError, raise_status
from .grpclib.status import Status, StatusCode
from purerpc.grpc_proto import GRPCProtoStream
from purerpc.grpclib.events import ResponseEnded


async def extract_message_from_singleton_stream(stream):
    msg = await stream.receive_message()
    if msg is None:
        event = stream.end_stream_event
        if isinstance(event, ResponseEnded):
            raise_status(event.status)
        raise ProtocolError("Expected one message, got zero")
    if await stream.receive_message() is not None:
        raise ProtocolError("Expected one message, got multiple")
    return msg


async def stream_to_async_iterator(stream: GRPCProtoStream):
    while True:
        msg = await stream.receive_message()
        if msg is None:
            event = stream.end_stream_event
            if isinstance(event, ResponseEnded):
                raise_status(event.status)
            return
        yield msg


async def send_multiple_messages_server(stream, agen):
    async with curio.meta.finalize(agen) as tmp:
        async for message in tmp:
            await stream.send_message(message)
    await stream.close(Status(StatusCode.OK))


async def send_single_message_server(stream, message):
    await stream.send_message(message)
    await stream.close(Status(StatusCode.OK))


async def send_multiple_messages_client(stream, agen):
    try:
        async with curio.meta.finalize(agen) as tmp:
            async for message in tmp:
                await stream.send_message(message)
    finally:
        await stream.close()


async def send_single_message_client(stream, message):
    await stream.send_message(message)
    await stream.close()


async def call_server_unary_unary(func, stream):
    msg = await extract_message_from_singleton_stream(stream)
    await send_single_message_server(stream, await func(msg))


async def call_server_unary_stream(func, stream):
    msg = await extract_message_from_singleton_stream(stream)
    await send_multiple_messages_server(stream, func(msg))


async def call_server_stream_unary(func, stream):
    input_message_stream = stream_to_async_iterator(stream)
    await send_single_message_server(stream, await func(input_message_stream))


async def call_server_stream_stream(func, stream):
    input_message_stream = stream_to_async_iterator(stream)
    await send_multiple_messages_server(stream, func(input_message_stream))


class ClientStub:
    def __init__(self, stream_fn):
        self._stream_fn = stream_fn


class ClientStubUnaryUnary(ClientStub):
    async def __call__(self, message):
        stream = await self._stream_fn()
        await send_single_message_client(stream, message)
        return await extract_message_from_singleton_stream(stream)


class ClientStubUnaryStream(ClientStub):
    async def __call__(self, message):
        stream = await self._stream_fn()
        await send_single_message_client(stream, message)
        async for message in stream_to_async_iterator(stream):
            yield message


class ClientStubStreamUnary(ClientStub):
    async def __call__(self, message_aiter):
        stream = await self._stream_fn()
        await curio.spawn(send_multiple_messages_client, stream, message_aiter, daemon=True)
        return await extract_message_from_singleton_stream(stream)


class ClientStubStreamStream(ClientStub):
    async def call_aiter(self, message_aiter):
        stream = await self._stream_fn()
        if message_aiter is not None:
            await curio.spawn(send_multiple_messages_client, stream, message_aiter, daemon=True)
            async for message in stream_to_async_iterator(stream):
                yield message

    async def call_stream(self):
        return await self._stream_fn()

    def __call__(self, message_aiter=None):
        if message_aiter is None:
            return self.call_stream()
        else:
            return self.call_aiter(message_aiter)



from .status import Status, StatusCode


class GRPCError(Exception):
    pass


class ProtocolError(GRPCError):
    pass


class MessageTooLargeError(ProtocolError):
	pass


class UnsupportedMessageEncodingError(ProtocolError):
    pass


class RpcFailedError(GRPCError):
	def __init__(self, status):
		super().__init__(f"RPC failed with status {status}")
		self._status = status

	@property
	def status(self):
		return self._status


class CancelledError(RpcFailedError):
	def __init__(self, message=""):
		super().__init__(Status(StatusCode.CANCELLED, message))


class UnknownError(RpcFailedError):
	def __init__(self, message=""):
		super().__init__(Status(StatusCode.UNKNOWN, message))


class InvalidArgumentError(RpcFailedError):
	def __init__(self, message=""):
		super().__init__(Status(StatusCode.INVALID_ARGUMENT, message))


class DeadlineExceededError(RpcFailedError):
	def __init__(self, message=""):
		super().__init__(Status(StatusCode.DEADLINE_EXCEEDED, message))


class NotFoundError(RpcFailedError):
	def __init__(self, message=""):
		super().__init__(Status(StatusCode.NOT_FOUND, message))


class AlreadyExistsError(RpcFailedError):
	def __init__(self, message=""):
		super().__init__(Status(StatusCode.ALREADY_EXISTS, message))


class PermissionDeniedError(RpcFailedError):
	def __init__(self, message=""):
		super().__init__(Status(StatusCode.PERMISSION_DENIED, message))


class ResourceExhaustedError(RpcFailedError):
	def __init__(self, message=""):
		super().__init__(Status(StatusCode.RESOURCE_EXHAUSTED, message))


class FailedPreconditionError(RpcFailedError):
	def __init__(self, message=""):
		super().__init__(Status(StatusCode.FAILED_PRECONDITION, message))


class AbortedError(RpcFailedError):
	def __init__(self, message=""):
		super().__init__(Status(StatusCode.ABORTED, message))


class OutOfRangeError(RpcFailedError):
	def __init__(self, message=""):
		super().__init__(Status(StatusCode.OUT_OF_RANGE, message))


class UnimplementedError(RpcFailedError):
	def __init__(self, message=""):
		super().__init__(Status(StatusCode.UNIMPLEMENTED, message))


class InternalError(RpcFailedError):
	def __init__(self, message=""):
		super().__init__(Status(StatusCode.INTERNAL, message))


class UnavailableError(RpcFailedError):
	def __init__(self, message=""):
		super().__init__(Status(StatusCode.UNAVAILABLE, message))


class DataLossError(RpcFailedError):
	def __init__(self, message=""):
		super().__init__(Status(StatusCode.DATA_LOSS, message))


class UnauthenticatedError(RpcFailedError):
	def __init__(self, message=""):
		super().__init__(Status(StatusCode.UNAUTHENTICATED, message))


def raise_status(status: Status):
	if status.status_code == StatusCode.CANCELLED:
		raise CancelledError(status.status_message)
	elif status.status_code == StatusCode.UNKNOWN:
		raise UnknownError(status.status_message)
	elif status.status_code == StatusCode.INVALID_ARGUMENT:
		raise InvalidArgumentError(status.status_message)
	elif status.status_code == StatusCode.DEADLINE_EXCEEDED:
		raise DeadlineExceededError(status.status_message)
	elif status.status_code == StatusCode.NOT_FOUND:
		raise NotFoundError(status.status_message)
	elif status.status_code == StatusCode.ALREADY_EXISTS:
		raise AlreadyExistsError(status.status_message)
	elif status.status_code == StatusCode.PERMISSION_DENIED:
		raise PermissionDeniedError(status.status_message)
	elif status.status_code == StatusCode.RESOURCE_EXHAUSTED:
		raise ResourceExhaustedError(status.status_message)
	elif status.status_code == StatusCode.FAILED_PRECONDITION:
		raise FailedPreconditionError(status.status_message)
	elif status.status_code == StatusCode.ABORTED:
		raise AbortedError(status.status_message)
	elif status.status_code == StatusCode.OUT_OF_RANGE:
		raise OutOfRangeError(status.status_message)
	elif status.status_code == StatusCode.UNIMPLEMENTED:
		raise UnimplementedError(status.status_message)
	elif status.status_code == StatusCode.INTERNAL:
		raise InternalError(status.status_message)
	elif status.status_code == StatusCode.UNAVAILABLE:
		raise UnavailableError(status.status_message)
	elif status.status_code == StatusCode.DATA_LOSS:
		raise DataLossError(status.status_message)
	elif status.status_code == StatusCode.UNAUTHENTICATED:
		raise UnauthenticatedError(status.status_message)



import logging
import datetime

import h2.stream
import h2.errors
import h2.events
import h2.connection
import h2.exceptions
from h2.settings import SettingCodes

from .status import Status
from .config import GRPCConfiguration
from .events import MessageReceived, RequestReceived, RequestEnded, ResponseReceived, ResponseEnded
from .exceptions import ProtocolError
from .buffers import MessageReadBuffer, MessageWriteBuffer

logger = logging.getLogger(__name__)


def monkey_patch_h2_reset_on_closed_stream_bug():
    """
    H2 raises h2.exceptions.StreamClosedError when receiving RST_STREAM on a closed stream that was
    not closed by endpoint, for example:
    
    -> HEADERS frame
    -> DATA frame with END_STREAM flag set  # stream is now half-closed (local)
    <- HEADERS frame
    <- DATA frame
    <- HEADERS frame with END_STREAM flag set  # stream is now closed
    <- RST_STREAM

    This is the case when calling unary-unary requests with purerpc client and grpcio server
    (it sends END_STREAM + RST_STREAM because the RPC call is unary-unary and it cannot receive
    any more data).

    To fix this, we use a workaround.
    """
    h2.stream._transitions[h2.stream.StreamState.CLOSED,
                           h2.stream.StreamInputs.RECV_RST_STREAM] = \
        (None, h2.stream.StreamState.CLOSED)

    h2.stream._transitions[h2.stream.StreamState.CLOSED,
                           h2.stream.StreamInputs.RECV_WINDOW_UPDATE] = \
        (None, h2.stream.StreamState.CLOSED)


monkey_patch_h2_reset_on_closed_stream_bug()


class GRPCConnection:
    MAX_HEADER_LIST_SIZE = 8192
    MAX_CONCURRENT_STREAMS = 2 ** 16
    MAX_INBOUND_FRAME_SIZE = 2 ** 24 - 1
    MAX_CONCURRENT_STREAMS = 1000

    def __init__(self, config: GRPCConfiguration):
        self.config = config
        self.h2_connection = h2.connection.H2Connection(config._h2_config)
        self._monkey_patch_h2_connection()
        self.message_read_buffers = {}

    def _monkey_patch_h2_connection(self):
        self.h2_connection.local_settings = h2.settings.Settings(
            client=self.config.client_side,
            initial_values={
                SettingCodes.MAX_CONCURRENT_STREAMS: self.MAX_CONCURRENT_STREAMS,
                SettingCodes.INITIAL_WINDOW_SIZE: 2 * self.config.max_message_length,
                SettingCodes.MAX_FRAME_SIZE: self.MAX_INBOUND_FRAME_SIZE,
                SettingCodes.MAX_HEADER_LIST_SIZE: self.MAX_HEADER_LIST_SIZE,
            }
        )
        self.h2_connection.max_inbound_frame_size = self.MAX_INBOUND_FRAME_SIZE

    def _request_received(self, event: h2.events.RequestReceived):
        if event.stream_ended:
            raise ProtocolError("Stream ended before data was sent")
        request = RequestReceived.parse_from_stream_id_and_headers_destructive(
            event.stream_id, dict(event.headers))
        self.message_read_buffers[event.stream_id] = MessageReadBuffer(request.message_encoding,
            self.config.max_message_length)
        return [request]

    def _response_received(self, event: h2.events.ResponseReceived):
        headers = dict(event.headers)
        response_received = ResponseReceived.parse_from_stream_id_and_headers_destructive(
            event.stream_id, headers)
        if event.stream_ended:
            response_ended = ResponseEnded.parse_from_stream_id_and_headers_destructive(
                event.stream_id, headers)
            return [response_received, response_ended]
        else:
            if len(headers) > 0:
                raise ProtocolError("Unparsed headers: {}".format(headers))
            self.message_read_buffers[event.stream_id] = MessageReadBuffer(
                response_received.message_encoding, self.config.max_message_length)
            return [response_received]

    def _trailers_received(self, event: h2.events.TrailersReceived):
        response_ended = ResponseEnded.parse_from_stream_id_and_headers_destructive(
            event.stream_id, dict(event.headers))
        return [response_ended]

    def _informational_response_received(self, event: h2.events.InformationalResponseReceived):
        return []

    def _data_received(self, event: h2.events.DataReceived):
        try:
            self.message_read_buffers[event.stream_id].data_received(event.data,
                                                                     event.flow_controlled_length)
        except KeyError:
            self.h2_connection.reset_stream(event.stream_id, h2.errors.ErrorCodes.PROTOCOL_ERROR)

        iterator = (self.message_read_buffers[event.stream_id]
                    .read_all_complete_messages_flowcontrol())
        events = []
        for message, flow_controlled_length in iterator:
            events.append(MessageReceived(event.stream_id, message, flow_controlled_length))
        return events

    def _window_updated(self, event: h2.events.WindowUpdated):
        return [event]

    def _remote_settings_changed(self, event: h2.events.RemoteSettingsChanged):
        fake_event = h2.events.WindowUpdated()
        fake_event.stream_id = 0
        fake_event.delta = 1
        return [fake_event]

    def _ping_acknowledged(self, event: h2.events.PingAcknowledged):
        return []

    def _stream_ended(self, event: h2.events.StreamEnded):
        if event.stream_id in self.message_read_buffers:
            del self.message_read_buffers[event.stream_id]
            return [RequestEnded(event.stream_id)] if not self.config.client_side else []
        return []

    def _stream_reset(self, event: h2.events.StreamReset):
        return []

    def _push_stream_received(self, event: h2.events.PushedStreamReceived):
        return []

    def _settings_acknowledged(self, event: h2.events.SettingsAcknowledged):
        # TODO: implement this
        return []

    def _priority_updated(self, event: h2.events.PriorityUpdated):
        return []

    def _connection_terminated(self, event: h2.events.ConnectionTerminated):
        return []

    def _alternative_service_available(self, event: h2.events.AlternativeServiceAvailable):
        return []

    def _unknown_frame_received(self, event: h2.events.UnknownFrameReceived):
        return []

    def initiate_connection(self):
        self.h2_connection.initiate_connection()
        self.h2_connection.increment_flow_control_window(2 ** 30)

    def data_to_send(self, amount: int = None):
        return self.h2_connection.data_to_send(amount)

    def receive_data(self, data: bytes):
        events = self.h2_connection.receive_data(data)
        grpc_events = []
        for event in events:
            if isinstance(event, h2.events.RequestReceived):
                grpc_events.extend(self._request_received(event))
            elif isinstance(event, h2.events.ResponseReceived):
                grpc_events.extend(self._response_received(event))
            elif isinstance(event, h2.events.TrailersReceived):
                grpc_events.extend(self._trailers_received(event))
            elif isinstance(event, h2.events.InformationalResponseReceived):
                grpc_events.extend(self._informational_response_received(event))
            elif isinstance(event, h2.events.DataReceived):
                grpc_events.extend(self._data_received(event))
            elif isinstance(event, h2.events.WindowUpdated):
                grpc_events.extend(self._window_updated(event))
            elif isinstance(event, h2.events.RemoteSettingsChanged):
                grpc_events.extend(self._remote_settings_changed(event))
            elif isinstance(event, h2.events.PingAcknowledged):
                grpc_events.extend(self._ping_acknowledged(event))
            elif isinstance(event, h2.events.StreamEnded):
                grpc_events.extend(self._stream_ended(event))
            elif isinstance(event, h2.events.StreamReset):
                grpc_events.extend(self._stream_reset(event))
            elif isinstance(event, h2.events.PushedStreamReceived):
                grpc_events.extend(self._push_stream_received(event))
            elif isinstance(event, h2.events.SettingsAcknowledged):
                grpc_events.extend(self._settings_acknowledged(event))
            elif isinstance(event, h2.events.PriorityUpdated):
                grpc_events.extend(self._priority_updated(event))
            elif isinstance(event, h2.events.ConnectionTerminated):
                grpc_events.extend(self._connection_terminated(event))
            elif isinstance(event, h2.events.AlternativeServiceAvailable):
                grpc_events.extend(self._alternative_service_available(event))
            elif isinstance(event, h2.events.UnknownFrameReceived):
                grpc_events.extend(self._unknown_frame_received(event))

        return grpc_events

    def flow_control_window(self, stream_id: int):
        return min(self.h2_connection.max_outbound_frame_size,
                   self.h2_connection.local_flow_control_window(stream_id))

    def reset_stream(self, stream_id: int, error_code: h2.errors.ErrorCodes):
        self.h2_connection.reset_stream(stream_id, error_code)

    def acknowledge_received_data(self, stream_id: int, flow_controlled_length: int):
        self.h2_connection.acknowledge_received_data(flow_controlled_length, stream_id)

    def send_data(self, stream_id: int, data: bytes, end_stream: bool = False):
        self.h2_connection.send_data(stream_id, data, end_stream=end_stream)

    def get_next_available_stream_id(self):
        return self.h2_connection.get_next_available_stream_id()

    def start_request(self, stream_id: int, scheme: str, service_name: str, method_name: str,
                      message_type=None, authority=None, timeout: datetime.timedelta=None,
                      content_type_suffix="", custom_metadata=()):
        headers = [
            (":method", "POST"),
            (":scheme", scheme),
            (":path", "/{}/{}".format(service_name, method_name)),
            ("te", "trailers"),
            ("content-type", "application/grpc" + content_type_suffix),
            *custom_metadata
        ]
        if authority is not None:
            headers.insert(3, (":authority", authority))
        if timeout is not None:
            number_of_whole_seconds = timeout.days * 86400 + timeout.seconds
            if timeout.microseconds == 0:
                timeout_unit = "S"
                timeout_value = number_of_whole_seconds
            else:
                timeout_unit = "u"
                timeout_value = number_of_whole_seconds * 1000000 + timeout.microseconds
            timeout_str = "{}{}".format(timeout_value, timeout_unit)
            headers.insert(4, ("grpc-timeout", timeout_str))
        if message_type is not None:
            headers.append(("grpc-message-type", message_type))
        if self.config._message_encoding is not None:
            headers.append(("grpc-encoding", self.config._message_encoding))
        if self.config._message_accept_encoding is not None:
            headers.append(("grpc-accept-encoding", self.config._message_accept_encoding))
        if self.config._user_agent is not None:
            headers.append(("user-agent", self.config._user_agent))
        self.h2_connection.send_headers(stream_id, headers, end_stream=False)

    def end_request(self, stream_id: int):
        self.h2_connection.send_data(stream_id, b"", end_stream=True)

    def start_response(self, stream_id: int, content_type_suffix="", custom_metadata=()):
        headers = [
            (":status", "200"),
            ("content-type", "application/grpc" + content_type_suffix),
            *custom_metadata,
        ]
        if self.config._message_encoding is not None:
            headers.append(("grpc-encoding", self.config._message_encoding))
        if self.config._message_accept_encoding is not None:
            headers.append(("grpc-accept-encoding", self.config._message_accept_encoding))
        self.h2_connection.send_headers(stream_id, headers, end_stream=False)

    def respond_status(self, stream_id: int, status: Status, content_type_suffix="",
                       custom_metadata=()):
        trailers = [
            (":status", "200"),
            ("content-type", "application/grpc" + content_type_suffix),
            ("grpc-status", str(status.int_value)),
            *custom_metadata,
        ]
        if status.status_message:
            # TODO: should be percent encoded
            trailers.append(("grpc-message", status.status_message))
        self.h2_connection.send_headers(stream_id, trailers, end_stream=True)

    def end_response(self, stream_id: int, status: Status, custom_metadata=()):
        trailers = [
            ("grpc-status", str(status.int_value)),
            *custom_metadata,
        ]
        if status.status_message:
            # TODO: should be percent encoded
            trailers.append(("grpc-message", status.status_message))
        self.h2_connection.send_headers(stream_id, trailers, end_stream=True)



from purerpc.client import Channel, Client
from purerpc.server import Service, Server
from purerpc.rpc import Cardinality, RPCSignature, Stream
from purerpc.grpclib.status import Status, StatusCode
from purerpc.grpclib.exceptions import *


import curio
from purerpc.grpc_proto import GRPCProtoSocket
import functools
from purerpc.grpclib.config import GRPCConfiguration
from purerpc.rpc import RPCSignature, Cardinality
from purerpc.wrappers import ClientStubUnaryUnary, ClientStubStreamStream, ClientStubUnaryStream, \
    ClientStubStreamUnary


class Channel:
    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.grpc_socket = None

    async def connect(self):
        socket = await curio.open_connection(self.host, self.port)
        config = GRPCConfiguration(client_side=True)
        self.grpc_socket = GRPCProtoSocket(config, socket)
        await self.grpc_socket.initiate_connection()


class Client:
    def __init__(self, service_name: str, channel: Channel):
        self.service_name = service_name
        self.channel = channel

    async def rpc(self, method_name: str, request_type, response_type):
        if self.channel.grpc_socket is None:
            await self.channel.connect()
        message_type = request_type.DESCRIPTOR.full_name
        stream = await self.channel.grpc_socket.start_request("http", self.service_name,
                                                              method_name, message_type,
                                                              "{}:{}".format(self.channel.host,
                                                                             self.channel.port))
        stream.expect_message_type(response_type)
        return stream

    def get_method_stub(self, method_name: str, signature: RPCSignature):
        stream_fn = functools.partial(self.rpc, method_name, signature.request_type,
                                      signature.response_type)
        if signature.cardinality == Cardinality.STREAM_STREAM:
            return ClientStubStreamStream(stream_fn)
        elif signature.cardinality == Cardinality.UNARY_STREAM:
            return ClientStubUnaryStream(stream_fn)
        elif signature.cardinality == Cardinality.STREAM_UNARY:
            return ClientStubStreamUnary(stream_fn)
        else:
            return ClientStubUnaryUnary(stream_fn)



"""
 Copyright 2014 gRPC authors.

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
import enum


class StatusCode(enum.Enum):
	# OK is returned on success.
	OK = 0

	# Canceled indicates the operation was canceled (typically by the caller).
	CANCELLED = 1

	# Unknown error. An example of where this error may be returned is
	# if a Status value received from another address space belongs to
	# an error-space that is not known in this address space. Also
	# errors raised by APIs that do not return enough error information
	# may be converted to this error.
	UNKNOWN = 2

	# InvalidArgument indicates client specified an invalid argument.
	# Note that this differs from FailedPrecondition. It indicates arguments
	# that are problematic regardless of the state of the system
	# (e.g., a malformed file name).
	INVALID_ARGUMENT = 3

	# DeadlineExceeded means operation expired before completion.
	# For operations that change the state of the system, this error may be
	# returned even if the operation has completed successfully. For
	# example, a successful response from a server could have been delayed
	# long enough for the deadline to expire.
	DEADLINE_EXCEEDED = 4

	# NotFound means some requested entity (e.g., file or directory) was
	# not found.
	NOT_FOUND = 5

	# AlreadyExists means an attempt to create an entity failed because one
	# already exists.
	ALREADY_EXISTS = 6

	# PermissionDenied indicates the caller does not have permission to
	# execute the specified operation. It must not be used for rejections
	# caused by exhausting some resource (use ResourceExhausted
	# instead for those errors). It must not be
	# used if the caller cannot be identified (use Unauthenticated
	# instead for those errors).
	PERMISSION_DENIED = 7

	# ResourceExhausted indicates some resource has been exhausted, perhaps
	# a per-user quota, or perhaps the entire file system is out of space.
	RESOURCE_EXHAUSTED = 8

	# FailedPrecondition indicates operation was rejected because the
	# system is not in a state required for the operation's execution.
	# For example, directory to be deleted may be non-empty, an rmdir
	# operation is applied to a non-directory, etc.
	#
	# A litmus test that may help a service implementor in deciding
	# between FailedPrecondition, Aborted, and Unavailable:
	#  (a) Use Unavailable if the client can retry just the failing call.
	#  (b) Use Aborted if the client should retry at a higher-level
	#      (e.g., restarting a read-modify-write sequence).
	#  (c) Use FailedPrecondition if the client should not retry until
	#      the system state has been explicitly fixed. E.g., if an "rmdir"
	#      fails because the directory is non-empty, FailedPrecondition
	#      should be returned since the client should not retry unless
	#      they have first fixed up the directory by deleting files from it.
	#  (d) Use FailedPrecondition if the client performs conditional
	#      REST Get/Update/Delete on a resource and the resource on the
	#      server does not match the condition. E.g., conflicting
	#      read-modify-write on the same resource.
	FAILED_PRECONDITION = 9

	# Aborted indicates the operation was aborted, typically due to a
	# concurrency issue like sequencer check failures, transaction aborts,
	# etc.
	#
	# See litmus test above for deciding between FailedPrecondition,
	# Aborted, and Unavailable.
	ABORTED = 10

	# OutOfRange means operation was attempted past the valid range.
	# E.g., seeking or reading past end of file.
	#
	# Unlike InvalidArgument, this error indicates a problem that may
	# be fixed if the system state changes. For example, a 32-bit file
	# system will generate InvalidArgument if asked to read at an
	# offset that is not in the range [0,2^32-1], but it will generate
	# OutOfRange if asked to read from an offset past the current
	# file size.
	#
	# There is a fair bit of overlap between FailedPrecondition and
	# OutOfRange. We recommend using OutOfRange (the more specific
	# error) when it applies so that callers who are iterating through
	# a space can easily look for an OutOfRange error to detect when
	# they are done.
	OUT_OF_RANGE = 11

	# Unimplemented indicates operation is not implemented or not
	# supported/enabled in this service.
	UNIMPLEMENTED = 12

	# Internal errors. Means some invariants expected by underlying
	# system has been broken. If you see one of these errors,
	# something is very broken.
	INTERNAL = 13

	# Unavailable indicates the service is currently unavailable.
	# This is a most likely a transient condition and may be corrected
	# by retrying with a backoff.
	#
	# See litmus test above for deciding between FailedPrecondition,
	# Aborted, and Unavailable.
	UNAVAILABLE = 14

	# DataLoss indicates unrecoverable data loss or corruption.
	DATA_LOSS = 15

	# Unauthenticated indicates the request does not have valid
	# authentication credentials for the operation.
	UNAUTHENTICATED = 16


class Status:
	def __init__(self, status_code_or_int, status_message=""):
		if isinstance(status_code_or_int, StatusCode):
			self._status_code = status_code_or_int
			self._int_value = status_code_or_int.value
		else:
			try:
				self._status_code = StatusCode(status_code_or_int)
			except ValueError:
				self._status_code = StatusCode.UNKNOWN
			self._int_value = status_code_or_int
		self._status_message = status_message

	@property
	def status_code(self):
		return self._status_code

	@property
	def int_value(self):
		return self._int_value

	@property
	def status_message(self):
		return self._status_message

	def __str__(self):
		suffix = ("" if not self.status_message else ": " + self.status_message)
		if self.status_code != StatusCode.UNKNOWN:
			prefix = self.status_code.name
		else:
			prefix = f"{self.status_code.name} ({self.int_value})"
		return prefix + suffix

	def __repr__(self):
		return (f"Status({self.status_code.name}, code={self.int_value}, "
			    f"message={repr(self.status_message)})")



from setuptools import setup, find_packages


def main():
    console_scripts = ['protoc-gen-purerpc=purerpc.protoc_plugin.plugin:main']
    setup(
        name="purerpc",
        version="0.1.0",
        author="Andrew Stepanov",
        description="Asynchronous pure python gRPC server and client implementation using curio "
                    "and hyper-h2.",
        packages=find_packages('src'),
        package_dir={'': 'src'},
        test_suite="tests",
        python_requires=">=3.6.0",
        install_requires=[
            "autopep8",
            "h2",
            "protobuf",
            "curio",
        ],
        entry_points={'console_scripts': console_scripts},
        setup_requires=["pytest-runner"],
        tests_require=[
            "pytest",
            "grpcio_tools",
            "grpcio"
        ]
    )


if __name__ == "__main__":
    main()



import sys
import inspect
import warnings
import collections
from multiprocessing import Process

import curio
import curio.meta
import typing
import logging

from .grpclib.status import Status, StatusCode
from .grpclib.exceptions import RpcFailedError
from purerpc.grpc_proto import GRPCProtoStream, GRPCProtoSocket
from purerpc.grpc_socket import GRPCSocket, GRPCStream
from purerpc.rpc import RPCSignature, Cardinality
from purerpc.utils import is_linux, get_linux_kernel_version
from purerpc.wrappers import stream_to_async_iterator, call_server_unary_unary, \
    call_server_unary_stream, call_server_stream_unary, call_server_stream_stream

from .grpclib.connection import GRPCConfiguration


BoundRPCMethod = collections.namedtuple("BoundRPCMethod", ["method_fn", "signature"])


class Service:
    def __init__(self, name):
        self.name = name
        self.methods = {}

    def add_method(self, method_name: str, method_fn, rpc_signature: RPCSignature):
        self.methods[method_name] = BoundRPCMethod(method_fn, rpc_signature)

    def rpc(self, method_name):
        def decorator(func):
            signature = inspect.signature(func)
            if signature.return_annotation == signature.empty:
                raise ValueError("Only annotated methods can be used with Service.rpc() decorator")
            if len(signature.parameters) != 1:
                raise ValueError("Only functions with one parameter can be used with Service.rpc("
                                 ") decorator")
            parameter = next(iter(signature.parameters.values()))
            if parameter.annotation == parameter.empty:
                raise ValueError("Only annotated methods can be used with Service.rpc() decorator")

            rpc_signature = RPCSignature.from_annotations(parameter.annotation,
                                                          signature.return_annotation)
            self.add_method(method_name, func, rpc_signature)
            return func
        return decorator


class Servicer:
    @property
    def service(self) -> Service:
        raise NotImplementedError()


class Server:
    def __init__(self, port=50055, num_processes=1):
        self.port = port
        self.services = {}
        self.num_processes = num_processes
        if num_processes > 1 and (not is_linux() or get_linux_kernel_version() < (3, 9)):
            warnings.warn("Selected num_processes > 1 and not running Linux kernel >= 3.9")

    def add_service(self, service):
        self.services[service.name] = service

    def _create_socket_and_listen(self):
        return curio.tcp_server_socket('', self.port, reuse_address=True, reuse_port=True)

    async def _run_async_server(self, socket):
        await curio.network.run_server(socket, lambda c, a: ConnectionHandler(self)(c, a))

    def _target_fn(self):
        socket = self._create_socket_and_listen()
        curio.run(self._run_async_server, socket)

    def serve(self):
        if self.num_processes == 1:
            self._target_fn()
        else:
            # this is simple SO_REUSEPORT load balancing on Linux
            processes = []
            for i in range(self.num_processes):
                process = Process(target=self._target_fn)
                process.start()
                processes.append(process)
            for process in processes:
                process.join()


class ConnectionHandler:
    RECEIVE_BUFFER_SIZE = 65536

    def __init__(self, server: Server):
        self.config = GRPCConfiguration(client_side=False)
        self.grpc_socket = None
        self.server = server

    async def request_received(self, stream: GRPCProtoStream):
        await stream.start_response()
        event = await stream.receive_event()

        try:
            service = self.server.services[event.service_name]
        except KeyError:
            await stream.close(Status(
                StatusCode.UNIMPLEMENTED,
                status_message=f"Service {event.service_name} is not implemented"
            ))
            return

        try:
            bound_rpc_method = service.methods[event.method_name]
        except KeyError:
            await stream.close(Status(
                StatusCode.UNIMPLEMENTED,
                status_message=f"Method {event.method_name} is not implemented in "
                               f"service {event.service_name}"
            ))
            return

        # TODO: Should at least pass through GeneratorExit
        try:
            method_fn = bound_rpc_method.method_fn
            cardinality = bound_rpc_method.signature.cardinality
            stream.expect_message_type(bound_rpc_method.signature.request_type)
            if cardinality == Cardinality.STREAM_STREAM:
                await call_server_stream_stream(method_fn, stream)
            elif cardinality == Cardinality.UNARY_STREAM:
                await call_server_unary_stream(method_fn, stream)
            elif cardinality == Cardinality.STREAM_UNARY:
                await call_server_stream_unary(method_fn, stream)
            else:
                await call_server_unary_unary(method_fn, stream)
        except RpcFailedError as error:
            await stream.close(error.status)
        except:
            logging.exception("Got exception while writing response stream")
            await stream.close(Status(StatusCode.CANCELLED, status_message=repr(sys.exc_info())))

    async def __call__(self, socket, addr):
        self.grpc_socket = GRPCProtoSocket(self.config, socket)
        await self.grpc_socket.initiate_connection()

        # TODO: TaskGroup() uses a lot of memory if the connection is kept for a long time
        # TODO: do we really need it here?
        # task_group = curio.TaskGroup()
        # TODO: Should at least pass through GeneratorExit
        try:
            async for stream in self.grpc_socket.listen():
                await curio.spawn(self.request_received(stream), daemon=True)
        except:
            logging.exception("Got exception in main dispatch loop")
        finally:
            # await task_group.join()
            # await self.grpc_socket.shutdown()
            pass



import curio
import socket
import curio.io
import datetime
import collections
import h2
import h2.events
from purerpc.grpclib.exceptions import ProtocolError

from .grpclib.connection import GRPCConfiguration, GRPCConnection
from .grpclib.events import RequestReceived, RequestEnded, ResponseEnded, MessageReceived
from .grpclib.buffers import MessageWriteBuffer, MessageReadBuffer


class SocketWrapper:
    def __init__(self, grpc_connection: GRPCConnection, sock: curio.io.Socket):
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        self._socket = sock
        self._grpc_connection = grpc_connection
        self._write_fifo_lock = curio.Lock()
        self._retry = False

    async def flush(self):
        """This maybe called from different threads."""
        async with self._write_fifo_lock:
            self._retry = True
            while self._retry:
                self._retry = False
                data = self._grpc_connection.data_to_send()
                if not data:
                    return
                await self._socket.sendall(data)

    async def try_flush(self):
        if self.locked:
            self._retry = True
        else:
            await self.flush()

    @property
    def locked(self):
        return self._write_fifo_lock.locked()

    async def recv(self, buffer_size: int):
        """This may only be called from single thread."""
        return await self._socket.recv(buffer_size)


class GRPCStream:
    def __init__(self, grpc_connection: GRPCConnection, stream_id: int, socket: SocketWrapper,
                 grpc_socket: "GRPCSocket"):
        self._stream_id = stream_id
        self._grpc_connection = grpc_connection
        self._grpc_socket = grpc_socket
        self._socket = socket
        self._flow_control_update_event = curio.Event()
        self._incoming_events = curio.Queue()

        self._response_started = False
        self._closed = False
        
        self._start_stream_event = None
        self._end_stream_event = None

    @property
    def start_stream_event(self):
        return self._start_stream_event

    @property
    def end_stream_event(self):
        return self._end_stream_event

    @property
    def flow_control_update_event(self):
        return self._flow_control_update_event

    @property
    def incoming_events(self):
        return self._incoming_events

    @property
    def stream_id(self):
        return self._stream_id

    @property
    def client_side(self):
        return self._grpc_connection.config.client_side

    @property
    def debug_prefix(self):
        return "[CLIENT] " if self.client_side else "[SERVER] "

    async def _send(self, message: bytes, compress=False):
        message_write_buffer = MessageWriteBuffer(self._grpc_connection.config.message_encoding,
                                                  self._grpc_connection.config.max_message_length)
        message_write_buffer.write_message(message, compress)
        while message_write_buffer:
            window_size = self._grpc_connection.flow_control_window(self._stream_id)
            if window_size <= 0:
                await self._flow_control_update_event.wait()
                self._flow_control_update_event.clear()
                continue
            num_data_to_send = min(window_size, len(message_write_buffer))
            data = message_write_buffer.data_to_send(num_data_to_send)
            self._grpc_connection.send_data(self._stream_id, data)
            await self._socket.flush()

    async def _receive(self):
        event = await self._incoming_events.get()
        if isinstance(event, MessageReceived):
            self._grpc_connection.acknowledge_received_data(self._stream_id,
                                                            event.flow_controlled_length)
            await self._socket.flush()
        elif isinstance(event, RequestEnded) or isinstance(event, ResponseEnded):
            assert self._end_stream_event is None
            self._end_stream_event = event
        else:
            assert self._start_stream_event is None
            self._start_stream_event = event
        return event

    async def close(self, status=None, content_type_suffix="", custom_metadata=()):
        if self.client_side and (status or custom_metadata):
            raise ValueError("Client side streams cannot be closed with non-default arguments")
        if self._closed:
            raise TypeError("Closing already closed stream")
        self._closed = True
        if self.client_side:
            self._grpc_connection.end_request(self._stream_id)
        elif self._response_started:
            self._grpc_connection.end_response(self._stream_id, status, custom_metadata)
        else:
            self._grpc_connection.respond_status(self._stream_id, status,
                                                 content_type_suffix, custom_metadata)
        await self._socket.flush()
        self._grpc_socket._decref_stream(self._stream_id)

    async def start_response(self, content_type_suffix="", custom_metadata=()):
        if self.client_side:
            raise ValueError("Cannot start response on client-side socket")
        self._grpc_connection.start_response(self._stream_id, content_type_suffix, custom_metadata)
        self._response_started = True
        await self._socket.flush()


# TODO: this name is not correct, should be something like GRPCConnection (but this name is already
# occupied)
class GRPCSocket:
    StreamClass = GRPCStream

    def __init__(self, config: GRPCConfiguration, sock: curio.io.Socket,
                 receive_buffer_size=16384):
        self._grpc_connection = GRPCConnection(config=config)
        self._socket = SocketWrapper(self._grpc_connection, sock)
        self._receive_buffer_size = receive_buffer_size

        self._streams = {}  # type: Dict[int, GRPCStream]
        self._streams_count = {}  # type: Dict[int, int]

    @property
    def client_side(self):
        return self._grpc_connection.config.client_side

    def _stream_ctor(self, stream_id):
        return self.StreamClass(self._grpc_connection, stream_id, self._socket, self)

    def _allocate_stream(self, stream_id):
        stream = self._stream_ctor(stream_id)
        self._streams[stream_id] = stream
        self._streams_count[stream_id] = 2
        return stream

    def _decref_stream(self, stream_id: int):
        self._streams_count[stream_id] -= 1
        if self._streams_count[stream_id] == 0:
            del self._streams[stream_id]
            del self._streams_count[stream_id]

    async def _listen(self):
        while True:
            data = await self._socket.recv(self._receive_buffer_size)
            if not data:
                return
            events = self._grpc_connection.receive_data(data)
            for event in events:
                # TODO: implement this
                if isinstance(event, h2.events.WindowUpdated):
                    if event.stream_id == 0:
                        for stream in self._streams.values():
                            await stream.flow_control_update_event.set()
                    elif event.stream_id in self._streams:
                        await self._streams[event.stream_id].flow_control_update_event.set()
                    continue
                elif isinstance(event, RequestReceived):
                    self._allocate_stream(event.stream_id)

                await self._streams[event.stream_id]._incoming_events.put(event)

                if isinstance(event, RequestReceived):
                    yield self._streams[event.stream_id]
                elif isinstance(event, ResponseEnded) or isinstance(event, RequestEnded):
                    self._decref_stream(event.stream_id)
            await self._socket.try_flush()

    async def _listener_thread(self):
        async for _ in self._listen():
            raise ProtocolError("Received request on client end")

    async def initiate_connection(self):
        self._grpc_connection.initiate_connection()
        await self._socket.flush()
        if self.client_side:
            await curio.spawn(self._listener_thread(), daemon=True)

    async def listen(self):
        if self.client_side:
            raise ValueError("Cannot listen client-side socket")
        async for stream in self._listen():
            yield stream

    async def start_request(self, scheme: str, service_name: str, method_name: str,
                            message_type=None, authority=None, timeout: datetime.timedelta=None,
                            content_type_suffix="", custom_metadata=()):
        if not self.client_side:
            raise ValueError("Cannot start request on server-side socket")
        stream_id = self._grpc_connection.get_next_available_stream_id()
        stream = self._allocate_stream(stream_id)
        self._grpc_connection.start_request(stream_id, scheme, service_name, method_name,
                                            message_type, authority, timeout,
                                            content_type_suffix, custom_metadata)
        await self._socket.flush()
        return stream
